package com.sales;


import org.apache.flink.api.common.functions.FlatJoinFunction;
import org.apache.flink.api.common.operators.Order;
import org.apache.flink.api.java.DataSet;
import org.apache.flink.api.java.ExecutionEnvironment;
import org.apache.flink.api.java.aggregation.Aggregations;
import org.apache.flink.api.java.tuple.Tuple2;
import org.apache.flink.api.java.tuple.Tuple5;
import org.apache.flink.api.java.tuple.Tuple7;
import org.apache.flink.util.Collector;

/**
 * Implements the sales program that gives insights about sales stats and performance report.
 */

public class SaleAnalysisJob {

	
	public static void main(String[] args) throws Exception {

		System.out.println("FolderPath: "+args[0]+" | FileNumber: "+args[1]);

//		set up the execution environment
		final ExecutionEnvironment env = ExecutionEnvironment.getExecutionEnvironment();

//		create customer dataset object from csv data file which have customer details
		DataSet<Customer> customerDataSet = env.readCsvFile(args[0]+"/customer.csv")
				.pojoType(Customer.class, "customer_id", "first_name", "last_name", "address", "city", "state", "zip", "email", "phone");

//		create product dataset object from csv data file which have product details
		DataSet<Product> productDataSet = env.readCsvFile(args[0]+"/product.csv")
				.pojoType(Product.class, "product_id", "name", "vendor", "catalog", "category", "current_price");

//		create sale customer dataset object from csv data file which have sales details along with customer details
		DataSet<Netsale_Customer> netsaleCustomerDataSet = env.readCsvFile(args[0]+"/netsale_customer_"+args[1]+".csv")
				.pojoType(Netsale_Customer.class, "netsale_id", "sale_time", "customer_id", "shipping", "tax", "total");

//		create sale product dataset object form csv data file
		DataSet<Netsale_Product> netsaleProductDataSet = env.readCsvFile(args[0]+"/netsale_product_"+args[1]+".csv")
				.pojoType(Netsale_Product.class, "netsale_id", "product_id", "price", "qty");

//		get top customer details
		getTopCustomers(customerDataSet, netsaleCustomerDataSet, args[0], args[1]);

//		get top product details
		getTopProducts(productDataSet, netsaleProductDataSet, args[0], args[1]);

//		get top customer product details
		getTopCustomerProductDetails(customerDataSet, netsaleCustomerDataSet, productDataSet, netsaleProductDataSet, args[0], args[1]);

		env.execute("SalesJob-"+args[1]);
	}

	/********************************************************************************
	 Top ten customers in terms of purchase value.
	 ********************************************************************************/
	static void getTopCustomers(DataSet<Customer> customerDataSet, DataSet<Netsale_Customer> netsaleCustomerDataSet, String folderPath, String fileNumber) {

//		Join sales and customer data and get collected values
		DataSet<Tuple7<Long, Long, String, String, String, Double, String>> saleCustomer = netsaleCustomerDataSet.join(customerDataSet)
				.where(netsaleCustomer -> netsaleCustomer.customer_id)
				.equalTo(customer -> customer.customer_id)
				.with(new FlatJoinFunction<Netsale_Customer, Customer, Tuple7<Long, Long, String, String, String, Double, String>>() {
					@Override
					public void join(Netsale_Customer netsaleCustomer, Customer customer, Collector<Tuple7<Long, Long, String, String, String, Double, String>> out) {
						out.collect(new Tuple7(netsaleCustomer.netsale_id, customer.customer_id, (customer.first_name+" "+customer.last_name), customer.state,
								customer.city, netsaleCustomer.total, netsaleCustomer.sale_time));
					}
				})
				;

//		group by customer sales data and sum purchase value and order by purchase value
		DataSet<Tuple7<Long, Long, String, String, String, Double, String>> saleCustomerSum = saleCustomer.groupBy(1)
				.aggregate(Aggregations.SUM, 5)
				.sortPartition(
						5, Order.DESCENDING
				);

//		get top 10 customer sales values
		saleCustomerSum.first(5).writeAsText(folderPath+"/topCustomers_"+fileNumber+".csv");

	}

	/********************************************************************************
	 Top ten products with the most purchase.
	 ********************************************************************************/
	static void getTopProducts(DataSet<Product> productDataSet, DataSet<Netsale_Product> netsaleProductDataSet, String folderPath, String fileNumber) {
//		Join sales and cutomer data and get collected values
		DataSet<Tuple7<Long, Long, String, String, String, String, Integer>> saleProduct = netsaleProductDataSet.join(productDataSet)
				.where(netsaleProduct -> netsaleProduct.product_id)
				.equalTo(product -> product.product_id)
				.with(new FlatJoinFunction<Netsale_Product, Product, Tuple7<Long, Long, String, String, String, String, Integer>>() {
					@Override
					public void join(Netsale_Product netsaleProduct, Product product, Collector<Tuple7<Long, Long, String, String, String, String, Integer>> out) {
						out.collect(new Tuple7(netsaleProduct.netsale_id, product.product_id, product.getName(), product.vendor, product.getCatalog(), product.getCategory(), netsaleProduct.qty));
					}
				})
				;

//		group by customer sales data and sum purchase value and order by purchase value
		DataSet<Tuple7<Long, Long, String, String, String, String, Integer>> saleProductSum = saleProduct.groupBy(2,3)
				.aggregate(Aggregations.SUM, 6)
				.sortPartition(
						6, Order.DESCENDING
				);

//		get top 10 customer sales values
		saleProductSum.first(5).writeAsText(folderPath+"/topProducts_"+fileNumber+".csv");
	}

	/********************************************************************************
	 Product details for top three customers in terms of purchase value.
	 ********************************************************************************/
	static void getTopCustomerProductDetails(DataSet<Customer> customerDataSet, DataSet<Netsale_Customer> netsaleCustomerDataSet,
									  DataSet<Product> productDataSet, DataSet<Netsale_Product> netsaleProductDataSet, String folderPath, String fileNumber) {
		DataSet<Tuple7<Long, Long, String, Double,  Long, String, Integer>> saleCustomerProduct =
				netsaleCustomerDataSet.join(customerDataSet)
						.where(netsaleCustomer -> netsaleCustomer.customer_id)
						.equalTo(customer -> customer.customer_id)

						.with(new FlatJoinFunction<Netsale_Customer, Customer, Tuple5<Long, Long, String, Double, String>>() {
							@Override
							public void join(Netsale_Customer netsaleCustomer, Customer customer, Collector<Tuple5<Long, Long, String, Double, String>> out) {
								out.collect(new Tuple5(netsaleCustomer.netsale_id, customer.customer_id, (customer.first_name+" "+customer.last_name), netsaleCustomer.total, netsaleCustomer.sale_time));
							}
						})
						.groupBy(1)
						.aggregate(Aggregations.SUM, 3)
						.sortPartition(3, Order.DESCENDING)
						.first(1)

						.join(netsaleProductDataSet)
						.where(netsaleCustomerDataSet1 -> netsaleCustomerDataSet1.f0)
						.equalTo(netsaleProductDataSet1 -> netsaleProductDataSet1.netsale_id)

						.join(productDataSet)
						.where(value -> value.f1.product_id)
						.equalTo(value -> value.product_id)

						.with(new FlatJoinFunction<Tuple2<Tuple5<Long, Long, String, Double, String>, Netsale_Product>, Product, Tuple7<Long, Long, String, Double, Long, String, Integer>>() {
							@Override
							public void join(Tuple2<Tuple5<Long, Long, String, Double, String>, Netsale_Product> tuple2, Product product, Collector<Tuple7<Long, Long, String, Double, Long, String, Integer>> out) {
								out.collect(new Tuple7(tuple2.f0.f0, tuple2.f0.f1, tuple2.f0.f2, tuple2.f0.f3, product.product_id, product.name, tuple2.f1.qty));
							}
						})
				;

		saleCustomerProduct.writeAsText(folderPath+"/topCustomerProductDetails_"+fileNumber+".csv");
	}

}
